#!/usr/bin/env python3

import yaml
import argparse
import Rx
import sys
import subprocess
import os
from string import Template

parser = argparse.ArgumentParser()
parser.add_argument("file", help="file to process", nargs='+')
parser.add_argument("--html", help="output block description as HTML",
                    action = 'store_true')
parser.add_argument("--latex", help="output block description as LaTeX",
                    action = 'store_true')
parser.add_argument("--latex-table", help="output block status table as LaTeX",
                    action = 'store_true')
args = parser.parse_args()

mydir = os.path.dirname(os.path.realpath(__file__))

def print_markdown_as(format, text):
    sys.stdout.flush()
    proc = subprocess.Popen(['pandoc', '-f', 'markdown', '-t', format], stdin = subprocess.PIPE)
    proc.stdin.write(bytes(text, 'UTF-8'))
    proc.communicate()
    proc.stdin.close()
    if proc.returncode:
        raise Exception("pandoc failed: %d" % proc.returncode)

def print_latex_desc(doc):
    def iodef_str(iodef):
        if iodef is None: return "None"
        str='%d\n\\begin{enumerate}\n' % len(iodef)
        for io in iodef:
            str += Template('\\item {\\bf $name} $type').substitute(io)
            if 'range' in io:
                str += '  %s' % io['range']
            str += '\n'
        return str +'\end{enumerate}'

    print("\\newpage\n")
    print("\\subsubsection{%s}\n" % doc['Name'])
    print("\\begin{description}\n")
    print("\\item[Inputs:]     %s\n" % iodef_str(doc['Inputs']))
    print("\\item[Outputs:]    %s\n" % iodef_str(doc['Outputs']))
    print("\\item[Parameters:] %s\n" % iodef_str(doc['Parameters']))
    print("\\end{description}\n")

    print_markdown_as('latex', doc['Description'])

    print("\n\\textbf{Status:}")
    print("\\begin{multicols}{3}")
    if doc['Status']['Tested']:
        print("\\begin{compactitem}")
        print("\\item \\textbf{Tested}:")
        print(" \\begin{compactitem}")
        print("\n".join(["\\item %s" % i.replace('_', '\\_') for i in doc['Status']['Tested']]))
        print(" \\end{compactitem}")
        print("\\end{compactitem}")
    else:
        print("\\ ")

    print("\\vfill\\columnbreak")

    if doc['Status']['Untested']:
        print("\\begin{compactitem}")
        print("\\item \\textbf{Untested}:")
        print(" \\begin{compactitem}")
        print("\n".join(["\\item %s" % i.replace('_', '\\_') for i in doc['Status']['Untested']]))
        print(" \\end{compactitem}")
        print("\\end{compactitem}")
    else:
        print("\\ ")

    print("\\vfill\\columnbreak")

    if doc['Status']['Not working']:
        print("\\begin{compactitem}")
        print("\\item \\textbf{Not working}:")
        print(" \\begin{compactitem}")
        print("\n".join(["\\item %s" % i.replace('_', '\\_') for i in doc['Status']['Not working']]))
        print(" \\end{compactitem}")
        print("\\end{compactitem}")
    else:
        print("\\ ")

    print("\\end{multicols}\n")

    if doc.get('RPP API functions used', None) is not None:
        print("\\textbf{RPP API functions used:}")
        print("\\begin{compactitem}")
        print("\n".join(["\\item \\texttt{%s}" % i.replace('_', '\\_') for i in doc['RPP API functions used']]))
        print("\\end{compactitem}")

    if doc.get('Relevant demos', None) is not None:
        print("\\textbf{Relevant demos:}")
        print("\\begin{compactitem}")
        print("\n".join(["\\item %s" % i.replace('_', '\\_') for i in doc['Relevant demos']]))
        print("\\end{compactitem}")

def process_file(f):
    yaml_doc = None
    for line in open(f, 'r'):
        if yaml_doc:
            yaml_doc += line
        elif line.startswith("%YAML"):
            yaml_doc = line
        if line == '...\n':
            break
    if yaml_doc is None:
        raise Exception("No line starting with '%%YAML' found in %s" % f)

    doc = yaml.load(yaml_doc)
    schema = yaml.load(open(mydir+'/schema.yaml', 'r'))

    rx = Rx.Factory({ "register_core_types": True })
    rpp_block_doc_schema = rx.make_schema(schema)
    try:
        rpp_block_doc_schema.check(doc)
    except Rx.ValidationError as e:
        raise Exception("Validation error in %s: %s" % (f, e))
        #raise

    if args.html:
        print_markdown_as('html', doc['Description'])
    if args.latex:
        print_latex_desc(doc)
    if args.latex_table:
        global last_category
        if last_category == doc['Category']: doc['Category']=''
        doc['Header'] = doc['Header'].replace('_', '\\_')
        if   doc['Status']['Not working']: doc['ShortStatus'] = '$\\alpha$'
        elif doc['Status']['Untested']:    doc['ShortStatus'] = '$\\beta$'
        elif not doc['Status']['Tested']:  doc['ShortStatus'] = '$\\beta$'
        else:                              doc['ShortStatus'] = 'Stable'
        print(Template("$Category & $Name & $ShortStatus & $Mnemonic & \\texttt{$Header} \\\\").substitute(doc))
        if doc['Category']:
            last_category = doc['Category']

last_category = None
for f in args.file:
    process_file(f)
