/* Copyright (C) 2014 Czech Technical University in Prague
 *
 * Authors:
 *     - Michal Horn <hornmich@fel.cvut.cz>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * File : sfunction_canrecive.c
 * Abstract:
 *     C-MEX S-function block for RPP CAN bus receive message.
 *
 * References:
 *     header.c
 *     trailer.c
 *
 * Compile with:
 *     mex(['-I' matlabroot '/toolbox/shared/can/src/scanutil'], ['-I' matlabroot '/toolbox/rtw/targets/common/can/datatypes'],  'sfunction_canreceive.c', [matlabroot '/toolbox/rtw/targets/common/can/datatypes/sfun_can_util.c'], [matlabroot '/toolbox/rtw/targets/common/can/datatypes/can_msg.c'])
 */

/*
%YAML 1.2
---
Name: CAN Receive
Category: CAN
Header: rpp/can.h
Mnemonic: CANR

Inputs:

Outputs:
  - { name: "f()", type: "Function call"                        }
  - { name: "Msg", type: "uint8, uint16, uint32 or CAN MESSAGE" }

Parameters:
  - { name: "Module",                   type: "Choice", range: "CAN1, CAN2, CAN3"                        }
  - { name: "Message ID type",          type: "Choice", range: "Standard, Extended, Mixed"               }
  - { name: "Message ID filter",        type: "Choice", range: "Single ID, ID and mask"                  }
  - { name: "Message ID",               type: "uint32"                                                   }
  - { name: "Message ID mask",          type: "uint8"                                                    }
  - { name: "Data type",                type: "Choice", range: "uint8, uint16, uint32, CAN MESSAGE TYPE" }
  - { name: "Automatic mailbox number", type: "bool"                                                     }
  - { name: "Mailbox number (0-31)",    type: "int8"                                                    }

# Description is in Markdown mark-up
Description: |

  This block allows receiving messages from the CAN bus. It can be
  configured for any of the CAN ports (modules) CAN1, CAN2 or CAN3.

  The acceptance rules for message reception can be specified by a
  *Message ID* parameter and optionally by a *Message ID mask*.
  Specifying the mask allows to receive messages with multiple IDs.
  The block supports both, the Standard (11b ID) and the Extended (29b
  ID) frame formats. Note that if Mixed message ID type is selected,
  the blocks will receive both frame types, but the Standard Message
  ID and optionally the Message ID mask has to be shifted by 18 bits
  to the left to correspond with the extended IDs and masks. For
  example, if Message ID parameter is set to 0x80000 and mask
  to 0x1ffbfff, the block will receive SFF messages with IDs 0x002 and
  0x003 and EFF IDs 0x00080000 and 0x000c0000.

  The mailbox number can be assigned automatically or manually. Automatic mailbox
  numbers are generated in ascending order from 0 to 31. Every mailbox must have a unique
  number. It is possible to mix blocks with automatically and manually
  assigned mailboxes. If the manually assigned mailbox number would
  collide with the automatic one then the automatically generated
  block will get assigned a next higher non-colliding ID.
  The mailbox numbers are shared between CAN Transmit and CAN Receive
  blocks with the same CAN port (module) parameter.

  The output of this block is a message data in selected format: uint8, uint16, uint32
  or CAN\_MESSAGE. The CAN\_MESSAGE object can be unpacked by `CAN Unpack` block.

  Every time a message is received, the function call on `f()` output
  signal is triggered and the received message data is appears on the
  `Msg` output port. See `cantransmit.slx` demo for examples of
  different configurations and the usage of the CAN blocks.

  In order to use this block, there must be a `CAN Configure` block in the model.

Status:
  Tested:
    - Reception of messages with a configured ID.
    - Rejection of messages with other ID then the configured one. FIXME
    - Reception of messages with a set of IDs specified by a ID and mask.
    - Reception of messages with Standard ID type, rejection of the messages with Extended ID type and vice versa.
    - Reception of messages with both messages ID types.
    - Automatic generation of mailboxes numbers with combination with manual specification, duplicate mailbox numbers detection.
    - Function call triggering when new message is accepted
  Untested:
    - Handling of error states on CAN bus
  Not working:
    - Receiving at baudrate higher than 700kb
    - External mode - throwing syntax error during compilation

RPP API functions used:
    - rpp_can_read()

Relevant demos:
    - cantransmit
    - can_demo
...
*/


#define S_FUNCTION_NAME sfunction_canreceive

#include "header.c"
#include <stdio.h>
#include "sfun_can_util.h"
#include "simstruc.h"

#define MSG_TYPE_STANDART_MAX	2048
#define MSG_TYPE_EXTENDED_MAX	536870912
#define MAILBOX_ID_MIN			1
#define MAILBOX_MAX_CNT			64

#define PARAM_NAME_MODULE_ID	"module_id"
#define PARAM_NAME_MAILBOX_ID	"mailbox_id"
#define PARAM_NAME_MSG_TYPE		"message_type"
#define PARAM_NAME_MSG_ID		"message_id"
#define PARAM_NAME_DATA_TYPE	"data_type"
#define PARAM_NAME_MSG_FILTER	"message_filter" 
#define PARAM_NAME_MSG_MASK		"message_mask"  
#define PARAM_NAME_MAILBOX_AUTO	"mailbox_auto"

/** Identifiers of the block parameters */
enum params{
	PARAM_MODULE_ID,
	PARAM_MAILBOX_ID,
	PARAM_MSG_TYPE,
	PARAM_MSG_ID,
	PARAM_DATA_TYPE,
	PARAM_MSG_FILTER,
	PARAM_MSG_MASK,
	PARAM_MAILBOX_AUTO,
	PARAM_COUNT
};

enum message_data_type {
	DATA_TYPE_UINT8 = 1,
	DATA_TYPE_UINT16 = 2,
	DATA_TYPE_UINT32 = 3,
	DATA_TYPE_CAN_MESSAGE = 4
};

enum message_id_type {
	MSG_ID_STANDART,
	MSG_ID_EXTENDED,
	MSG_ID_MIXED
};

enum outputs {
	OUT_FNC_CALL,
	OUT_MSG,
	OUT_COUNT
};

static bool msg_received;

static void mdlInitializeSizes(SimStruct *S){

	CAN_Common_MdlInitSizes(S);

	/*DTypeId type;*/

	if(!rppSetNumParams(S, PARAM_COUNT)) {
		return;
	}

	/* No input ports */
	if(!ssSetNumInputPorts(S, 0)) {
		return;
	}
     
	/*
	 * Configure output ports: 1
	 *	- Received Message
	 *	- Message is received
	 */
	if(!ssSetNumOutputPorts(S, OUT_COUNT)) {
		return;
	}
	rppAddOutputPort(S, OUT_FNC_CALL, SS_FCN_CALL);

	if ((int_T)mxGetPr(ssGetSFcnParam(S, PARAM_DATA_TYPE))[0] == DATA_TYPE_UINT8) {
		rppAddOutputPort(S, OUT_MSG, SS_UINT8);
	}
	else if ((int_T)mxGetPr(ssGetSFcnParam(S, PARAM_DATA_TYPE))[0] == DATA_TYPE_UINT16) {
		rppAddOutputPort(S, OUT_MSG, SS_UINT16);
	}
	else if ((int_T)mxGetPr(ssGetSFcnParam(S, PARAM_DATA_TYPE))[0] == DATA_TYPE_UINT32) {
		rppAddOutputPort(S, OUT_MSG, SS_UINT32);
	}
	else {
		rppAddOutputPort(S, OUT_MSG, ssGetDataTypeId(S, "CAN_MESSAGE"));
	}

	/* Set standard options for this block */
	rppSetStandardOptions(S);
}


#ifdef MATLAB_MEX_FILE
#define MDL_CHECK_PARAMETERS

static void mdlCheckParameters(SimStruct *S){
    /* Check the parameter mailbox number */
	if ((int_T)mxGetPr(ssGetSFcnParam(S, PARAM_MAILBOX_AUTO))[0] == 0) {
		if (!rppValidParamRange(S, PARAM_MAILBOX_ID, MAILBOX_ID_MIN, MAILBOX_ID_MIN + MAILBOX_MAX_CNT)) {
			return;
		}
	}

	int min = 0;
	int max = 0;
	if ((int_T)mxGetPr(ssGetSFcnParam(S, PARAM_MSG_TYPE))[0] == MSG_ID_STANDART) {
		max = MSG_TYPE_STANDART_MAX;
	}
	else {
		max = MSG_TYPE_EXTENDED_MAX;
	}

    /* Check the parameter message identifier */
    if (!rppValidParamRange(S, PARAM_MSG_ID, min, max)) {
        return;
    }
}
#endif

static void mdlInitializeSampleTimes(SimStruct *S)
{
	ssSetSampleTime(S, 0, INHERITED_SAMPLE_TIME);
	ssSetOffsetTime(S, 0, 0);

    ssSetCallSystemOutput(S,0);  /* call on first element */
}


#ifdef MATLAB_MEX_FILE
#define MDL_SET_WORK_WIDTHS
static void mdlSetWorkWidths(SimStruct *S){

	if(!ssSetNumRunTimeParams(S, PARAM_COUNT)) {
    	return;
	}

	ssRegDlgParamAsRunTimeParam(S, PARAM_MODULE_ID,  PARAM_MODULE_ID, PARAM_NAME_MODULE_ID, SS_UINT8);
	ssRegDlgParamAsRunTimeParam(S, PARAM_MAILBOX_ID,  PARAM_MAILBOX_ID, PARAM_NAME_MAILBOX_ID, SS_INT8);
	ssRegDlgParamAsRunTimeParam(S, PARAM_MSG_TYPE,  PARAM_MSG_TYPE, PARAM_NAME_MSG_TYPE, SS_UINT16);   
	ssRegDlgParamAsRunTimeParam(S, PARAM_MSG_ID,  PARAM_MSG_ID, PARAM_NAME_MSG_ID, SS_UINT32); 
  	ssRegDlgParamAsRunTimeParam(S, PARAM_DATA_TYPE,  PARAM_DATA_TYPE, PARAM_NAME_DATA_TYPE, SS_UINT8);   
  	ssRegDlgParamAsRunTimeParam(S, PARAM_MSG_FILTER,  PARAM_MSG_FILTER, PARAM_NAME_MSG_FILTER, SS_UINT8);   
  	ssRegDlgParamAsRunTimeParam(S, PARAM_MSG_MASK,  PARAM_MSG_MASK, PARAM_NAME_MSG_MASK, SS_UINT32);   
  	ssRegDlgParamAsRunTimeParam(S, PARAM_MAILBOX_AUTO,  PARAM_MAILBOX_AUTO, PARAM_NAME_MAILBOX_AUTO, SS_BOOLEAN);   
}
#endif

#define UNUSED_MDLOUTPUTS
#define UNUSED_MDLTERMINATE
#include "trailer.c"
